/*
 *
 *	This file is part of GLogiK project.
 *	GLogiK, daemon to handle special features on gaming keyboards
 *	Copyright (C) 2016-2018  Fabrice Delliaux <netbox253@gmail.com>
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <errno.h>

#include <cstring>
#include <cstdint>
#include <unistd.h>

#include <set>
#include <sstream>

#include <sys/inotify.h>

#include <config.h>


#define UTILS_COMPILATION 1

#include "log.h"
#include "exception.h"

#undef UTILS_COMPILATION

#include "FSNotify.h"


namespace NSGKUtils
{

FSNotify::FSNotify()
	:	inotify_queue_fd_(0)
{
	this->inotify_queue_fd_ = inotify_init1(IN_NONBLOCK);
	if( this->inotify_queue_fd_ == -1 ) {
		std::ostringstream buffer("inotify init failure : ", std::ios_base::app);
		buffer << strerror(errno);
		throw GLogiKExcept( buffer.str() );
	}
}

FSNotify::~FSNotify() {
	/* sanity check */
	if( ! this->watch_descriptors_.empty() ) {
		LOG(WARNING) << "some watch descriptors were not removed";
		for(const auto & watched_pair : this->watch_descriptors_) {
			this->removeNotifyWatch(watched_pair.second.wd);
		}
		this->watch_descriptors_.clear();
	}

	/* closing queue */
	if(this->inotify_queue_fd_ > -1) {
		if( close(this->inotify_queue_fd_) == -1 ) {
			LOG(ERROR) << "inotify queue closing failure : " << strerror(errno);
		}
	}
}

/* make sure path is readable before calling this one */
const int FSNotify::addNotifyDirectoryWatch(const std::string & path, const bool check_if_already_watched) {
	/* check first if this path is already watched */
	/* returns its descriptor in this case */
	if( check_if_already_watched ) {
		auto it = this->watch_descriptors_.find(path);
		if ( it != this->watch_descriptors_.end() ) {
#if DEBUGGING_ON
			LOG(DEBUG2) << "path already watched - " << path;
#endif
			return (*it).second.wd;
		}
	}

	uint32_t mask = IN_MODIFY | IN_DELETE_SELF | IN_MOVE_SELF | IN_ONLYDIR;
	return this->addNotifyWatch(path, mask);
}

void FSNotify::removeNotifyWatch(const int wd) {
	if(wd == -1) {
		LOG(ERROR) << "[" << wd << "] - negative watch descriptor";
		return;
	}

	const int & fd = this->inotify_queue_fd_;
	auto find_descriptor = [&fd, &wd] (auto & item) -> const bool {
		watchedObject & watched = item.second;
		if(watched.wd != wd)
			return false;
		--watched.count;
		if(watched.count == 0) {
			if( inotify_rm_watch(fd, wd) == -1 ) {
				LOG(ERROR) << "inotify rm watch failure : " << strerror(errno);
			}
			else {
#if DEBUGGING_ON
				LOG(DEBUG2) << "[" << wd << "] - removed path notify - " << item.first;
#endif
			}
			return true;
		}
#if DEBUGGING_ON
		LOG(DEBUG2) << "[" << wd << "] - path reference decremented - " << item.first;
#endif
		return false;
	};

	erase_if(this->watch_descriptors_, find_descriptor);
}

const int FSNotify::getNotifyQueueDescriptor(void) const {
	return this->inotify_queue_fd_;
}

void FSNotify::readNotifyEvents(files_map_t & files_map) {
	char *ptr;
	ssize_t len;
	char buf[4096] __attribute__ ((aligned(__alignof__(struct inotify_event))));
	const struct inotify_event *event;

	std::set<std::string> to_remove;

	for (;;) {
		/* Read some events. */
		len = read(this->inotify_queue_fd_, buf, sizeof buf);
		if (len == -1 && errno != EAGAIN) {
			throw GLogiKExcept("read error");
		}

		if (len <= 0)
			break;

		for (ptr = buf; ptr < buf + len; ptr += sizeof(struct inotify_event) + event->len) {
			event = (const struct inotify_event *) ptr;

			std::string wd("[");
			wd += std::to_string(event->wd);
			wd += "]";

			/* file inside watched directory */
			if (event->len) {
				const std::string name(event->name);
#if DEBUGGING_ON
				LOG(DEBUG2) << wd << " - name: " << name;
#endif
				/* map size will be 0 or 1 after that */
				erase_if(files_map, [&name]( auto & item ) -> const bool { return item.second != name; } );
			}
			else { /* watched object event */
				files_map.clear();

				std::string path;
				for(const auto & item_pair : this->watch_descriptors_) {
					if(item_pair.second.wd == event->wd) {
						path = item_pair.first;
						break;
					}
				}
				if( path.empty() ) {
					/*
					 * on next readNotifyEvents call IN_IGNORED can be generated by
					 * current IN_MOVE_SELF since we force remove the watch below
					 */
					if( event->mask & IN_IGNORED ) {
#if DEBUGGING_ON
						LOG(DEBUG2) << wd << "[IN_IGNORED] watch was removed";
#endif
					}
					else {
#if DEBUGGING_ON
						LOG(DEBUG2) << wd << "skip event : path not found";
#endif
					}
					continue; /* next event */
				}

				if( event->mask & IN_MOVE_SELF ) {
#if DEBUGGING_ON
					LOG(DEBUG2) << wd << "[IN_MOVE_SELF] watched object renamed or moved : " << path;
#endif
					to_remove.insert(path);
				}
				if( event->mask & IN_DELETE_SELF ) {
#if DEBUGGING_ON
					LOG(DEBUG2) << wd << "[IN_DELETE_SELF] watched object deleted : " << path;
#endif
				}
				if( event->mask & IN_IGNORED ) {
#if DEBUGGING_ON
					LOG(DEBUG2) << wd << "[IN_IGNORED] watch was removed : " << path;
#endif
					this->watch_descriptors_.erase(path);
				}
			}
		}

		for(const auto & path : to_remove) {
			auto it = this->watch_descriptors_.find(path);
			if(it != this->watch_descriptors_.end()) {
				(*it).second.count = 1; /* force removing watch */
				this->removeNotifyWatch((*it).second.wd);
			}
		}
		to_remove.clear();
	}
}

/*
 * --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
 * --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
 *
 * === private === private === private === private === private ===
 *
 * --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
 * --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
 */

const int FSNotify::addNotifyWatch(const std::string & path, const uint32_t & mask) {
	auto it = this->watch_descriptors_.find(path);

	if(it == this->watch_descriptors_.end()) {
		const int ret = inotify_add_watch(this->inotify_queue_fd_, path.c_str(), mask);
		if(ret < 0) {
			std::ostringstream buffer("inotify path watch failure : ", std::ios_base::app);
			buffer << path << " - " << strerror(errno);
			throw GLogiKExcept( buffer.str() );
		}

		this->watch_descriptors_.insert( std::pair<const std::string, watchedObject>(path, watchedObject(ret)) );
		it = this->watch_descriptors_.find(path);

#if DEBUGGING_ON
		LOG(DEBUG2) << "[" << (*it).second.wd << "] - added path notify - " << path;
#endif
	}
	else {
		(*it).second.count++;

#if DEBUGGING_ON
		LOG(DEBUG2) << "[" << (*it).second.wd << "] - path reference incremented - " << path;
#endif
	}

	return (*it).second.wd;
}


} // namespace NSGKUtils

